from datetime import datetime
from .app_config import AppConfig

class JobScheduler:

    def __init__(self):
        self.app_config = AppConfig()

    @classmethod
    def exit_handler(cls, scheduler, debug):
        try:
            # Remove job and shut down the scheduler
            scheduler.remove_job('moesif_events_batch_job')
            scheduler.shutdown()
        except Exception as ex:
            if debug:
                print("Error while closing the queue or scheduler shut down")
                print(str(ex))

    def send_events(self, api_client, batch_events, debug):
        try:
            if debug:
                print("Sending events to Moesif")
            batch_events_api_response = api_client.create_events_batch(batch_events)
            if debug:
                print("Events sent successfully")
            # Fetch Config ETag from response header
            batch_events_response_config_etag = batch_events_api_response.get("X-Moesif-Config-ETag")
            batch_events_response_rule_etag = batch_events_api_response.get("X-Moesif-Rules-Tag", None)
            # Return Config Etag
            return batch_events_response_config_etag, batch_events_response_rule_etag
        except Exception as ex:
            if debug:
                print("Error sending event to Moesif")
                print(str(ex))
            return None

    # Function to fetch application config
    def fetch_app_config(self, config, config_etag, sampling_percentage, last_updated_time, api_client, debug):
        try:
            config = self.app_config.get_config(api_client, debug)
            if config:
                config_etag, sampling_percentage, last_updated_time = self.app_config.parse_configuration(config, debug)
        except Exception as e:
            if debug:
                print('Error while fetching the application configuration')
                print(str(e))
        return config, config_etag, sampling_percentage, last_updated_time


    def batch_events(self, api_client, moesif_events_queue, debug, batch_size):

        batch_events = []
        try:
            while not moesif_events_queue.empty():
                batch_events.append(moesif_events_queue.get_nowait())
                if len(batch_events) == batch_size:
                    break

            if batch_events:
                batch_response_config_etag, batch_response_rules_etag = self.send_events(api_client, batch_events, debug)
                batch_events[:] = []
                # Set the last time event job ran after sending events
                return batch_response_config_etag, batch_response_rules_etag, datetime.utcnow()
            else:
                if debug:
                    print("No events to send")
                # Set the last time event job ran but no message to read from the queue
                return None, None, datetime.utcnow()
        except Exception as e:
            if debug:
                print("No message to read from the queue")
                print(str(e))
            # Set the last time event job ran when exception occurred while sending event
            return None, None, datetime.utcnow()



from .governance_rule_response import GovernanceRuleBlockResponse


class BlockResponseBufferList:
    def __init__(self):
        self.responses = []
        self.rule_type = None
        self.blocked = False

    def update(self, block, updated_gr_status, updated_gr_headers, updated_gr_body):
        if not self.rule_type:
            self.rule_type = 'regex'
        if block and not self.blocked:
            self.blocked = True

        gov_rule_response = GovernanceRuleBlockResponse()
        gov_rule_response.update_response(updated_gr_status, updated_gr_headers, updated_gr_body, block)
        self.responses.append(gov_rule_response)



from __future__ import print_function

import requests
import threading
import random
import math
import queue
import logging
from django.conf import settings
from django.utils import timezone
from moesifapi.moesif_api_client import *
from moesifapi.api_helper import *
from moesifapi.exceptions.api_exception import *
from moesifapi.models import *
from django.http import HttpRequest, HttpResponse
from .http_response_catcher import HttpResponseCatcher
from .masks import MaskData
from .client_ip import *
from .update_users import *
from .update_companies import *
from io import BytesIO
from moesifpythonrequest.start_capture.start_capture import StartCapture
from datetime import datetime, timedelta
from .app_config import AppConfig
from .parse_body import ParseBody
from .update_companies import Company
from .update_users import User
from .client_ip import ClientIp
from .logger_helper import LoggerHelper
from .event_mapper import EventMapper
from .job_scheduler import JobScheduler
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.interval import IntervalTrigger
from apscheduler.events import EVENT_JOB_ERROR, EVENT_JOB_EXECUTED
import atexit

class MoesifMiddlewarePre19(object):

    # One-time configuration and initialization.
    def __init__(self):
        self.middleware_settings = settings.MOESIF_MIDDLEWARE
        self.DEBUG = self.middleware_settings.get('LOCAL_DEBUG', False)
        self.LOG_BODY = self.middleware_settings.get('LOG_BODY', True)
        self.client = MoesifAPIClient(self.middleware_settings.get('APPLICATION_ID'))
        # below comment for setting moesif base_uri to a test server.
        if self.middleware_settings.get('LOCAL_DEBUG', False):
            Configuration.BASE_URI = self.get_configuration_uri(self.middleware_settings, 'BASE_URI', 'LOCAL_MOESIF_BASEURL')
        Configuration.version = 'moesifdjango-python/2.1.5'
        if settings.MOESIF_MIDDLEWARE.get('CAPTURE_OUTGOING_REQUESTS', False):
            try:
                if self.DEBUG:
                    print('Start capturing outgoing requests')
                # Start capturing outgoing requests
                StartCapture().start_capture_outgoing(settings.MOESIF_MIDDLEWARE)
            except:
                print('Error while starting to capture the outgoing events')
        self.api_version = self.middleware_settings.get('API_VERSION')
        self.api_client = self.client.api
        response_catcher = HttpResponseCatcher()
        self.api_client.http_call_back = response_catcher
        self.app_config = AppConfig()
        self.parse_body = ParseBody()
        self.client_ip = ClientIp()
        self.logger_helper = LoggerHelper()
        self.event_mapper = EventMapper()
        self.job_scheduler = JobScheduler()
        self.mask_helper = MaskData()
        self.user = User()
        self.company = Company()
        self.config = self.app_config.get_config(self.api_client, self.DEBUG)
        self.sampling_percentage = 100
        self.config_etag = None
        self.last_updated_time = datetime.utcnow()
        self.last_event_job_run_time = datetime(1970, 1, 1, 0, 0)  # Assuming job never ran, set it to epoch start time
        self.scheduler = None
        self.event_queue_size = self.middleware_settings.get('EVENT_QUEUE_SIZE', 10000)
        self.mo_events_queue = queue.Queue(maxsize=self.event_queue_size)
        self.event_batch_size = self.middleware_settings.get('BATCH_SIZE', 25)
        self.is_event_job_scheduled = False
        try:
            if self.config:
                self.config_etag, self.sampling_percentage, self.last_updated_time = self.app_config.parse_configuration(
                    self.config, self.DEBUG)
        except:
            if self.DEBUG:
                print('Error while parsing application configuration on initialization')
        self.transaction_id = None

    # Function to get configuration uri
    def get_configuration_uri(self, settings, field, deprecated_field):
        uri = settings.get(field)
        if uri:
            return uri
        else:
            return settings.get(deprecated_field, 'https://api.moesif.net')

    # Function to listen to the send event job response
    def event_listener(self, event):
        if event.exception:
            if self.DEBUG:
                print('Error reading response from the scheduled event job')
        else:
            if event.retval:
                response_etag, self.last_event_job_run_time = event.retval
                if response_etag is not None \
                        and self.config_etag is not None \
                        and self.config_etag != response_etag \
                        and datetime.utcnow() > self.last_updated_time + timedelta(minutes=5):
                    try:
                        self.config, self.config_etag, self.sampling_percentage, self.last_updated_time = \
                            self.job_scheduler.fetch_app_config(self.config, self.config_etag, self.sampling_percentage,
                                                                self.last_updated_time, self.api_client, self.DEBUG)
                    except Exception as ex:
                        if self.DEBUG:
                            print('Error while updating the application configuration')
                            print(str(ex))

    # Function to schedule send event job in async
    def schedule_event_background_job(self):
        try:
            if not self.scheduler:
                self.scheduler = BackgroundScheduler(daemon=True)
            if not self.scheduler.get_jobs():
                self.scheduler.add_listener(self.event_listener, EVENT_JOB_EXECUTED | EVENT_JOB_ERROR)
                self.scheduler.start()
                self.scheduler.add_job(
                    func=lambda: self.job_scheduler.batch_events(self.api_client, self.mo_events_queue, self.DEBUG,
                                                                 self.event_batch_size),
                    trigger=IntervalTrigger(seconds=2),
                    id='moesif_events_batch_job',
                    name='Schedule events batch job every 2 second',
                    replace_existing=True)

                # Avoid passing logging message to the ancestor loggers
                logging.getLogger('apscheduler.executors.default').setLevel(logging.WARNING)
                logging.getLogger('apscheduler.executors.default').propagate = False

                # Exit handler when exiting the app
                atexit.register(lambda: self.job_scheduler.exit_handler(self.scheduler, self.DEBUG))
        except Exception as ex:
            if self.DEBUG:
                print("Error when scheduling the job")
                print(str(ex))

    @classmethod
    def process_request(cls, request):
        request.moesif_req_time = datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3]
        try:
            request._mo_body = request.body
            request._stream = BytesIO(request.body)
            request._read_started = False
        except:
            request._mo_body = None

    def process_response(self, request, response):
        # Code to be executed for each request before
        # the view (and later middleware) are called.

        req_time = request.moesif_req_time

        if self.DEBUG:
            print("raw body before getting response")

        # response = get_response(request)
        # Code to be executed for each request/response after
        # the view is called.

        # Check if need to skip logging event
        skip_event_response = self.logger_helper.skip_event(request, response, self.middleware_settings, self.DEBUG)
        if skip_event_response:
            return skip_event_response
        # Request headers
        req_headers = self.logger_helper.parse_request_headers(request, self.middleware_settings, self.DEBUG)

        # Add transaction id to request headers
        req_headers, self.transaction_id = self.logger_helper.add_transaction_id_to_header(req_headers, self.transaction_id,
                                                                                      self.middleware_settings)

        # Prepare Request Body
        req_body, req_body_transfer_encoding = self.logger_helper.prepare_request_body(request, req_headers,
                                                                                       self.LOG_BODY,
                                                                                       self.middleware_settings)
        # Fetch Ip Address
        ip_address = self.client_ip.get_client_ip(request)

        # Request URI
        uri = self.logger_helper.request_url(request)

        # Add transaction id to request headers
        if self.transaction_id:
            try:
                response._headers["x-moesif-transaction-id"] = ("X-Moesif-Transaction-Id", self.transaction_id)
            except AttributeError:
                response.headers["x-moesif-transaction-id"] = self.transaction_id

        # Parse Response headers
        rsp_headers = self.logger_helper.parse_response_headers(response, self.middleware_settings)

        # Prepare Response Body
        rsp_body, rsp_body_transfer_encoding = self.logger_helper.prepare_response_body(response, rsp_headers,
                                                                                        self.LOG_BODY,
                                                                                        self.middleware_settings)

        # Response Time
        rsp_time = datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3]

        # Prepare Event Request Model
        event_req = self.event_mapper.to_request(req_time, uri, request.method, self.api_version,
                                                 ip_address, req_headers, req_body, req_body_transfer_encoding)

        # Prepare Event Response Model
        event_rsp = self.event_mapper.to_response(rsp_time, response.status_code, rsp_headers, rsp_body,
                                                  rsp_body_transfer_encoding)

        # User Id
        user_id = self.logger_helper.get_user_id(self.middleware_settings, request, response, req_headers, self.DEBUG)

        # Company Id
        company_id = self.logger_helper.get_company_id(self.middleware_settings, request, response, self.DEBUG)

        # Event Metadata
        metadata = self.logger_helper.get_metadata(self.middleware_settings, request, response, self.DEBUG)

        # Session Token
        session_token = self.logger_helper.get_session_token(self.middleware_settings, request, response, self.DEBUG)

        # Prepare Event Model
        event_model = self.event_mapper.to_event(event_req, event_rsp, user_id, company_id, session_token, metadata)

        # Mask Event Model
        event_model = self.logger_helper.mask_event(event_model, self.middleware_settings, self.DEBUG)

        # Create random percentage
        random_percentage = random.random() * 100
        self.sampling_percentage = self.app_config.get_sampling_percentage(self.config, user_id, company_id)
        if self.sampling_percentage >= random_percentage:
            event_model.weight = 1 if self.sampling_percentage == 0 else math.floor(100 / self.sampling_percentage)
            try:
                if not self.is_event_job_scheduled and datetime.utcnow() > self.last_event_job_run_time + timedelta(
                        minutes=5):
                    try:
                        self.schedule_event_background_job()
                        self.is_event_job_scheduled = True
                        self.last_event_job_run_time = datetime.utcnow()
                    except Exception as ex:
                        self.is_event_job_scheduled = False
                        if self.DEBUG:
                            print('Error while starting the event scheduler job in background')
                            print(str(ex))
                if self.DEBUG:
                    print("Add Event to the queue")
                self.mo_events_queue.put(event_model)
            except Exception as ex:
                if self.DEBUG:
                    print("Error while adding event to the queue")
                    print(str(ex))

        return response

    def update_user(self, user_profile):
        self.user.update_user(user_profile, self.api_client, self.DEBUG)

    def update_users_batch(self, user_profiles):
        self.user.update_users_batch(user_profiles, self.api_client, self.DEBUG)

    def update_company(self, company_profile):
        self.company.update_company(company_profile, self.api_client, self.DEBUG)

    def update_companies_batch(self, companies_profiles):
        self.company.update_companies_batch(companies_profiles, self.api_client, self.DEBUG)



from datetime import datetime
import json
from moesifapi.exceptions.api_exception import *

# Application Configuration
class AppConfig:

    def __init__(self):
        pass

    @classmethod
    def get_config(cls, api_client, debug):
        """Get Config"""
        try:
            config_api_response = api_client.get_app_config()
            return config_api_response
        except APIException as inst:
            if 401 <= inst.response_code <= 403:
                print("Unauthorized access getting application configuration. Please check your Appplication Id.")
            if debug:
                print("Error getting application configuration, with status code:")
                print(inst.response_code)
        except Exception as ex:
            if debug:
                print("Error getting application configuration:")
                print(str(ex))

    @classmethod
    def parse_configuration(cls, config, debug):
        """Parse configuration object and return Etag, sample rate and last updated time"""
        try:
            return config.headers.get("X-Moesif-Config-ETag"), json.loads(config.raw_body).get('sample_rate', 100), datetime.utcnow()
        except:
            if debug:
                print('Error while parsing the configuration object, setting the sample rate to default')
            return None, 100, datetime.utcnow()

    @classmethod
    def get_rules_etag(cls, config, debug):
        """Get moesif rules etag"""
        try:
            return config.headers.get("x-moesif-rules-tag")
        except Exception as e:
            if debug:
                print('Error while parsing rules etag: ', e)
            return None


    @classmethod
    def get_sampling_percentage(cls, config, user_id, company_id):
        """Get sampling percentage"""

        if config is not None:
            try:
                config_body = json.loads(config.raw_body)

                user_sample_rate = config_body.get('user_sample_rate', None)

                company_sample_rate = config_body.get('company_sample_rate', None)

                if user_id and user_sample_rate and user_id in user_sample_rate:
                    return user_sample_rate[user_id]

                if company_id and company_sample_rate and company_id in company_sample_rate:
                    return company_sample_rate[company_id]

                return config_body.get('sample_rate', 100)
            except Exception as e:
                print("Error while parsing user or company sample rate")
                print(e)

        # Use default
        return 100



from moesifapi.models import *


class EventMapper:

    def __init__(self):
        pass

    @classmethod
    def to_request(self, req_time, uri, method, api_version, ip_address, req_headers, req_body, transfer_encoding):
        return EventRequestModel(time=req_time,
                                 uri=uri,
                                 verb=method,
                                 api_version=api_version,
                                 ip_address=ip_address,
                                 headers=req_headers,
                                 body=req_body,
                                 transfer_encoding=transfer_encoding)

    @classmethod
    def to_response(self, rsp_time, status_code, rsp_headers, rsp_body, transfer_encoding):
        return EventResponseModel(time=rsp_time,
                                  status=status_code,
                                  headers=rsp_headers,
                                  body=rsp_body,
                                  transfer_encoding=transfer_encoding)

    @classmethod
    def to_event(cls, event_req, event_rsp, user_id, company_id, session_token, metadata):
        # Prepare Event Model
        return EventModel(request=event_req,
                          response=event_rsp,
                          user_id=user_id,
                          company_id=company_id,
                          session_token=session_token,
                          metadata=metadata,
                          direction="Incoming")



class GovernanceRuleBlockResponse:
    # TODO Create a empty init, and add function
    def __init__(self):
        self.block_response_status = None
        self.block_response_headers = None
        self.block_response_body = None
        self.blocked = False

    def update_response(self, gr_status, gr_headers, gr_body, block):
        self.block_response_status = gr_status
        self.block_response_headers = gr_headers
        self.block_response_body = gr_body
        self.blocked = block



from moesifapi.models import *
from moesifapi.exceptions.api_exception import *
from moesifapi.api_helper import *


class Company:

    def __init__(self):
        pass

    @classmethod
    def update_company(cls, company_profile, api_client, DEBUG):
        if not company_profile:
            print('Expecting the input to be either of the type - CompanyModel, dict or json while updating user')
        else:
            if isinstance(company_profile, dict):
                if 'company_id' in company_profile:
                    try:
                        api_client.update_company(CompanyModel.from_dictionary(company_profile))
                        if DEBUG:
                            print('Company Profile updated successfully')
                    except APIException as inst:
                        if 401 <= inst.response_code <= 403:
                            print("Unauthorized access sending event to Moesif. Please check your Appplication Id.")
                        if DEBUG:
                            print("Error while updating company, with status code:")
                            print(inst.response_code)
                else:
                    print('To update a company, a company_id field is required')

            elif isinstance(company_profile, CompanyModel):
                if company_profile.company_id is not None:
                    try:
                        api_client.update_company(company_profile)
                        if DEBUG:
                            print('Company Profile updated successfully')
                    except APIException as inst:
                        if 401 <= inst.response_code <= 403:
                            print("Unauthorized access sending event to Moesif. Please check your Appplication Id.")
                        if DEBUG:
                            print("Error while updating company, with status code:")
                            print(inst.response_code)
                else:
                    print('To update a company, a company_id field is required')
            else:
                try:
                    company_profile_json = APIHelper.json_deserialize(company_profile)
                    if 'company_id' in company_profile_json:
                        try:
                            api_client.update_company(CompanyModel.from_dictionary(company_profile_json))
                            if DEBUG:
                                print('Company Profile updated successfully')
                        except APIException as inst:
                            if 401 <= inst.response_code <= 403:
                                print("Unauthorized access sending event to Moesif. Please check your Appplication Id.")
                            if DEBUG:
                                print("Error while updating company, with status code:")
                                print(inst.response_code)
                    else:
                        print('To update a company, a company_id field is required')
                except:
                    print('Error while deserializing the json, please make sure the json is valid')

    @classmethod
    def update_companies_batch(cls, companies_profiles, api_client, DEBUG):
        if not companies_profiles:
            print('Expecting the input to be either of the type - List of CompanyModel, dict or json while updating users')
        else:
            if all(isinstance(company, dict) for company in companies_profiles):
                if all('company_id' in company for company in companies_profiles):
                    try:
                        batch_profiles = [CompanyModel.from_dictionary(d) for d in companies_profiles]
                        api_client.update_companies_batch(batch_profiles)
                        if DEBUG:
                            print('Companies Profile updated successfully')
                    except APIException as inst:
                        if 401 <= inst.response_code <= 403:
                            print("Unauthorized access sending event to Moesif. Please check your Appplication Id.")
                        if DEBUG:
                            print("Error while updating companies, with status code:")
                            print(inst.response_code)
                else:
                    print('To update companies, an company_id field is required')

            elif all(isinstance(company, CompanyModel) for company in companies_profiles):
                if all(company.company_id is not None for company in companies_profiles):
                    try:
                        api_client.update_companies_batch(companies_profiles)
                        if DEBUG:
                            print('Companies Profile updated successfully')
                    except APIException as inst:
                        if 401 <= inst.response_code <= 403:
                            print("Unauthorized access sending event to Moesif. Please check your Appplication Id.")
                        if DEBUG:
                            print("Error while updating companues, with status code:")
                            print(inst.response_code)
                else:
                    print('To update companies, a company_id field is required')
            else:
                try:
                    company_profiles_json = [APIHelper.json_deserialize(d) for d in companies_profiles]
                    if all(isinstance(company, dict) for company in company_profiles_json) and all(
                                    'company_id' in company for company in company_profiles_json):
                        try:
                            batch_profiles = [CompanyModel.from_dictionary(d) for d in company_profiles_json]
                            api_client.update_companies_batch(batch_profiles)
                            if DEBUG:
                                print('Companies Profile updated successfully')
                        except APIException as inst:
                            if 401 <= inst.response_code <= 403:
                                print("Unauthorized access sending event to Moesif. Please check your Appplication Id.")
                            if DEBUG:
                                print("Error while updating companies, with status code:")
                                print(inst.response_code)
                    else:
                        print('To update companies, an company_id field is required')
                except:
                    print('Error while deserializing the json, please make sure the json is valid')



from moesifapi.api_helper import *
from .masks import MaskData
from .parse_body import ParseBody
from django.http import HttpRequest, HttpResponse
import json
import base64
import re
import copy
import uuid


class LoggerHelper:
    def __init__(self):
        self.regex_http_ = re.compile(r'^HTTP_.+$')
        self.regex_content_type = re.compile(r'^CONTENT_TYPE$')
        self.regex_content_length = re.compile(r'^CONTENT_LENGTH$')
        self.mask_helper = MaskData()
        self.parse_body = ParseBody()

    @classmethod
    def flatten_to_string(cls, value):
        if type(value) == str:
            return value
        if value is None:
            return ''
        return APIHelper.json_serialize(value)

    @classmethod
    def mapper(cls, response, key):
        return copy.deepcopy(response[key])

    def parse_request_headers(self, request, middleware_settings, debug):
        req_headers = {}
        regex_http_start = re.compile('^HTTP_')
        try:
            for header in request.META:
                if self.regex_http_.match(header) or self.regex_content_type.match(
                        header) or self.regex_content_length.match(header):
                    normalized_header = regex_http_start.sub('', header)
                    normalized_header = normalized_header.replace('_', '-')
                    req_headers[normalized_header] = request.META[header]
            req_headers = self.mask_helper.mask_headers(req_headers, middleware_settings.get('REQUEST_HEADER_MASKS'))
        except Exception as inst:
            if debug:
                print("error encountered while copying request header")
                print(inst)
            req_headers = {}

        if debug:
            print("about to print what is in meta %d " % len(request.META))
            for x in request.META:
                print(x, ':', request.META[x])
            print("about to print headers %d " % len(req_headers))
            for x in req_headers:
                print(x, ':', req_headers[x])

        req_headers = {k: self.flatten_to_string(v) for k, v in req_headers.items()}
        return req_headers

    def prepare_request_body(self, request, req_headers, log_body, middleware_settings):
        req_body = None
        req_body_transfer_encoding = None
        if log_body and request._mo_body:
            if isinstance(request._mo_body, str):
                req_body, req_body_transfer_encoding = self.parse_body.parse_string_body(request._mo_body,
                                                                                         self.parse_body.transform_headers(
                                                                                             req_headers),
                                                                                         middleware_settings.get(
                                                                                             'REQUEST_BODY_MASKS'))
            else:
                req_body, req_body_transfer_encoding = self.parse_body.parse_bytes_body(request._mo_body,
                                                                                        self.parse_body.transform_headers(
                                                                                            req_headers),
                                                                                        middleware_settings.get(
                                                                                            'REQUEST_BODY_MASKS'))
        return req_body, req_body_transfer_encoding

    def parse_response_headers(self, response, middleware_settings):
        # a little hacky, using _headers, which is intended as a private variable.
        try:
            rsp_headers = {k: self.mapper(response, k) for k, v in response._headers.items()}
        except AttributeError:
            rsp_headers = {k: self.mapper(response, k) for k, v in response.headers.items()}
        return self.mask_helper.mask_headers(rsp_headers, middleware_settings.get('RESPONSE_HEADER_MASKS'))

    def prepare_response_body(self, response, rsp_headers, log_body, middleware_settings):
        rsp_body = None
        rsp_body_transfer_encoding = None
        if log_body and isinstance(response, HttpResponse) and response.content:
            if isinstance(response.content, str):
                rsp_body, rsp_body_transfer_encoding = self.parse_body.parse_string_body(response.content,
                                                                                         self.parse_body.transform_headers(
                                                                                             rsp_headers),
                                                                                         middleware_settings.get(
                                                                                             'RESPONSE_BODY_MASKS'))
            else:
                rsp_body, rsp_body_transfer_encoding = self.parse_body.parse_bytes_body(response.content,
                                                                                        self.parse_body.transform_headers(
                                                                                            rsp_headers),
                                                                                        middleware_settings.get(
                                                                                            'RESPONSE_BODY_MASKS'))
        return rsp_body, rsp_body_transfer_encoding

    @classmethod
    def request_url(cls, request):
        return request.scheme + "://" + request.get_host() + request.get_full_path()

    @classmethod
    def add_transaction_id_to_header(cls, req_headers, transaction_id, middleware_settings):
        capture_transaction_id = middleware_settings.get('DISABLE_TRANSACTION_ID', False)
        if not capture_transaction_id:
            req_trans_id = req_headers.get("X-MOESIF-TRANSACTION-ID", None)
            if req_trans_id:
                transaction_id = req_trans_id
                if not transaction_id:
                    transaction_id = str(uuid.uuid4())
            else:
                transaction_id = str(uuid.uuid4())
            # Add transaction id to the request header
            req_headers["X-MOESIF-TRANSACTION-ID"] = transaction_id
        return req_headers, transaction_id

    @classmethod
    def transform_token(cls, token):
        if not isinstance(token, str):
            token = token.decode('utf-8')
        return token

    @classmethod
    def fetch_token(cls, token, token_type):
        return token.split(token_type, 1)[1].strip()

    @classmethod
    def split_token(cls, token):
        return token.split('.')

    def parse_authorization_header(self, token, field, debug):
        try:
            # Fix the padding issue before decoding
            token += '=' * (-len(token) % 4)
            # Decode the payload
            base64_decode = base64.b64decode(token)
            # Transform token to string to be compatible with Python 2 and 3
            base64_decode = self.transform_token(base64_decode)
            # Convert the payload to json
            json_decode = json.loads(base64_decode)
            # Convert keys to lowercase
            json_decode = {k.lower(): v for k, v in json_decode.items()}
            # Check if field is present in the body
            if field in json_decode:
                # Fetch user Id
                return str(json_decode[field])
        except Exception as e:
            if debug:
                print("Error while parsing authorization header to fetch user id.")
                print(e)
        return None

    def get_user_id(self, middleware_settings, request, response, request_headers, debug):
        user_id = None
        try:
            identify_user = middleware_settings.get('IDENTIFY_USER', None)
            if identify_user is not None:
                user_id = identify_user(request, response)
            if not user_id:
                if request.user and request.user.is_authenticated and request.user.username:
                    return request.user.username
                else:
                    # Transform request headers keys to lower case
                    request_headers = {k.lower(): v for k, v in request_headers.items()}
                    # Fetch the auth header name from the config
                    auth_header_names = middleware_settings.get('AUTHORIZATION_HEADER_NAME', 'authorization').lower()
                    # Split authorization header name by comma
                    auth_header_names = [x.strip() for x in auth_header_names.split(',')]
                    # Fetch the header name available in the request header
                    token = None
                    for auth_name in auth_header_names:
                        # Check if the auth header name in request headers
                        if auth_name in request_headers:
                            # Fetch the token from the request headers
                            token = request_headers[auth_name]
                            # Split the token by comma
                            token = [x.strip() for x in token.split(',')]
                            # Fetch the first available header
                            if len(token) >= 1:
                                token = token[0]
                            else:
                                token = None
                            break
                    # Fetch the field from the config
                    field = middleware_settings.get('AUTHORIZATION_USER_ID_FIELD', 'sub').lower()
                    # Check if the auth header name in request headers
                    if token:
                        # Check if token is of type Bearer
                        if 'Bearer' in token:
                            # Fetch the bearer token
                            token = self.fetch_token(token, 'Bearer')
                            # Split the bearer token by dot(.)
                            split_token = self.split_token(token)
                            # Check if payload is not None
                            if len(split_token) >= 3 and split_token[1]:
                                # Parse and set user Id
                                user_id = self.parse_authorization_header(split_token[1], field, debug)
                        # Check if token is of type Basic
                        elif 'Basic' in token:
                            # Fetch the basic token
                            token = self.fetch_token(token, 'Basic')
                            # Decode the token
                            decoded_token = base64.b64decode(token)
                            # Transform token to string to be compatible with Python 2 and 3
                            decoded_token = self.transform_token(decoded_token)
                            # Fetch and set the user Id
                            user_id = decoded_token.split(':', 1)[0].strip()
                        # Check if token is of user-defined custom type
                        else:
                            # Split the token by dot(.)
                            split_token = self.split_token(token)
                            # Check if payload is not None
                            if len(split_token) > 1 and split_token[1]:
                                # Parse and set user Id
                                user_id = self.parse_authorization_header(split_token[1], field, debug)
                            else:
                                # Parse and set user Id
                                user_id = self.parse_authorization_header(token, field, debug)
        except Exception as e:
            if debug:
                print("can not execute identify_user function, please check moesif settings.")
                print(e)
        return user_id

    @classmethod
    def get_company_id(cls, middleware_settings, request, response, debug):
        company_id = None
        try:
            identify_company = middleware_settings.get('IDENTIFY_COMPANY', None)
            if identify_company is not None:
                company_id = identify_company(request, response)
        except:
            if debug:
                print("can not execute identify_company function, please check moesif settings.")
        return company_id

    @classmethod
    def get_metadata(cls, middleware_settings, request, response, debug):
        metadata = None
        try:
            get_metadata = middleware_settings.get('GET_METADATA', None)
            if get_metadata is not None:
                metadata = get_metadata(request, response)
        except:
            if debug:
                print("can not execute get_metadata function, please check moesif settings.")
        return metadata

    @classmethod
    def get_session_token(cls, middleware_settings, request, response, debug):
        session_token = None
        try:
            if request.session and request.session.session_key:
                session_token = request.session.session_key
            else:
                get_session_token = middleware_settings.get('GET_SESSION_TOKEN', None)
                if get_session_token is not None:
                    session_token = get_session_token(request, response)
        except:
            if debug:
                print("Can not execute get_session_token function. Please check moesif settings.")
        return session_token

    @classmethod
    def skip_event(cls, request, response, middleware_settings, debug):
        try:
            skip_event = middleware_settings.get('SKIP', None)
            if skip_event is not None:
                if skip_event(request, response):
                    return response
        except:
            if debug:
                print("Having difficulty executing skip_event function. Please check moesif settings.")
            return None

    @classmethod
    def mask_event(cls, event_model, middleware_settings, debug):
        try:
            mask_event_model = middleware_settings.get('MASK_EVENT_MODEL', None)
            if mask_event_model is not None:
                event_model = mask_event_model(event_model)
        except:
            if debug:
                print("Can not execute mask_event_model function. Please check moesif settings.")
        return event_model



import base64

from django.conf import settings
import json
import re
from .block_response_buffer import BlockResponseBufferList
from .event_mapper import *
from .governance_rule_response import GovernanceRuleBlockResponse

REVERSED_PRIORITY_RULES_ORDER = ['regex', 'company', 'user']


class MoesifGovRuleHelper:
    def __init__(self):
        pass

    @classmethod
    def fetch_entity_rules_from_app_config(cls, config, debug):
        """
        get fetch entity rules from config
        :param config:
        :return: None, if there is no rules in config (no dynamic variable set with gov rules with the app)
                Otherwise, return mapping with keys user_rules and company_rules
        """
        entity_rules = {'user_rules': {}, 'company_rules': {}}
        try:
            raw_body = json.loads(config.raw_body)

            entity_rules = {
                'user_rules': raw_body.get('user_rules', {}),
                'company_rules': raw_body.get('company_rules', {}),
            }
            if debug:
                print(
                    f"[moesif] config got {len(entity_rules['user_rules'])} cohort users and {len(entity_rules['company_rules'])} cohort companies")

        except Exception as e:
            print("[moesif] Error when fetching raw_body from config, ", e)

        return entity_rules

    @classmethod
    def get_entity_governance_rule_and_check_block(cls, entity_rules, governance_rules):
        """
        fetch governance rule object from governance rules list based on entity rules and return the block value (True or False)
        :param entity_rules:
        :param governance_rules:
        :return:
        """
        block_entity_rules = []
        for rules in entity_rules:
            if 'rules' in rules:
                rule_id = rules['rules']
                if rule_id and rule_id in governance_rules:
                    if governance_rules[rule_id]['block']:
                        # if the rule is block checked, then add to the list
                        block_entity_rules.append(rules)
        return block_entity_rules

    @classmethod
    def fetch_governance_rule_response_details(cls, governance_rule):
        # Response status
        status = governance_rule['response']['status']
        # Response headers
        header = governance_rule['response']['headers']
        body = None
        if 'body' in governance_rule['response']:
            body = governance_rule['response']['body']
        return status, header, body

    def transform_values(self, data, rule_values):
        """
        Replace body value in the response body for the short-circuited request
        :param data:
        :param rule_values:
        :return:
        """
        if data is None:
            return None
        if isinstance(data, str):
            max_index = max(rule_values.keys())

            rule_values_list = [rule_values[key] if key in rule_values.keys() else None for key in
                                range(max_index + 1)]

            try:
                data = data.format(*rule_values_list).format(*rule_values_list)
                return data
            except Exception as e:
                return '[moesif] response body, to get the governance rule response, please try again later'
        if isinstance(data, dict):
            for k, v in data.items():
                data[k] = self.transform_values(v, rule_values)
            return data
        if isinstance(data, list) or isinstance(data, tuple):
            for index in len(data):
                data[index] = self.transform_values(data[index], rule_values)
            return data
        return data

    def regex_pattern_match(self, event_value, condition_value):
        """
        Perform the regex pattern matching with event value and condition value
        :param event_value:
        :param condition_value:
        :return:
        """
        return re.fullmatch(condition_value, event_value) is not None

    # remove the sample rate return value
    def check_request_with_regex_match(self, gr_regex_configs_list, request_mapping_for_regex_config,
                                       ready_for_body_request):
        """
        fetch the sample rate and determine if request is matched with the regex
        :param gr_regex_configs_list:
        :param request_mapping_for_regex_config:
        :param ready_for_body_request:
        :return:
        """
        for or_group_of_regex_rule in gr_regex_configs_list:
            # Fetch the conditions
            conditions = or_group_of_regex_rule.get('conditions', {})
            # fetch the sample rate
            # sample_rate = or_group_of_regex_rule.get('sample_rate', 100)

            # Bool flag to determine if the regex conditions are matched
            regex_matched = None
            # Create a dict to hold the conditions mapping (path and value)
            condition_path_value_mapping = {}

            for condition in conditions:
                try:
                    if 'request.body.' in condition['path']:
                        ok_for_request_body_regex = self.check_if_condition_for_request_body_field(condition)

                        # Add condition path -> value to the condition table
                        if not ok_for_request_body_regex:
                            continue
                    condition_path_value_mapping[condition['path']] = condition['value']

                except KeyError as ke:
                    print('Error when fetching info from regex config each condition: ', ke)
                except Exception as e:
                    print(e)

            # Iterate through conditions table and perform `and` operation between each conditions
            for path, values in condition_path_value_mapping.items():
                try:
                    if re.fullmatch('request\.body\..*', path):
                        if not ready_for_body_request:
                            continue
                except Exception as e:
                    print("[moesif] Error when matching path starts with request\\.body\\", e)

                # Check if the path exists in the request config mapping
                if path in request_mapping_for_regex_config:
                    try:
                        # Fetch the value of the path in request config mapping
                        event_data = request_mapping_for_regex_config[path]
                        # Perform regex matching with event value
                        regex_matched = self.regex_pattern_match(event_data, values)
                    except Exception as e:
                        print("[moesif] Error while matching condition of governance rule {} and event data {}".format(values, e))
                else:
                    # Path does not exist in request config mapping, so no need to match regex condition rule
                    regex_matched = False

                # If one of the rule does not match, skip the condition and avoid matching other rules for the same condition
                if not regex_matched:
                    break

            # If regex conditions matched, return sample rate and block request (true)
            if regex_matched:
                return True

        # If regex conditions are not matched, return default sample rate (nil) and do not block request (false)
        return False

    def check_event_matched_with_governance_rules(self, gr_regex_configs, request_mapping_for_regex_config,
                                                  ready_for_body_request):
        """
        check if the request config mapping governance rule regex conditions
        :param gr_regex_configs:
        :param request_mapping_for_regex_config:
        :param ready_for_body_request:
        :return:
        """
        matched = self.check_request_with_regex_match(gr_regex_configs,
                                                 request_mapping_for_regex_config,
                                                 ready_for_body_request)

        return matched

    @classmethod
    def get_req_content_type(cls, request):
        """
        fetch 'content-type' from request headers
        :param request:
        :return:
        """
        lower_case_key_headers =  {k.lower(): v for k, v in request.headers.items()}
        return lower_case_key_headers.get('content-type')

    def ok_request_body_regex_rule(self, request, req_body_transfer_encoding):
        """
        Check if the request body is ready for request.body regex rule
        request.headers.content-type contains “json” and the request body is a JSON object
        (or string/byte buffer that can be parsed as an object)
        :param request:
        :param req_body_transfer_encoding:
        :return: True if we get the special case of request content-type contains grarphql
        and the request body is in type of string
        """
        if not request.body or not req_body_transfer_encoding:
            return False

        content_type_value = self.get_req_content_type(request)
        if isinstance(request.body, str):
            if content_type_value == 'application/graphql':
                return True

        # request.headers.content-type contains “json” and the request body is a JSON object
        # (or string/byte buffer that can be parsed as an object)
        if 'json' not in content_type_value and req_body_transfer_encoding != 'json':
            return False

        return True

    @classmethod
    def check_if_condition_for_request_body_field(cls, condition):
        # check if The field value must be a JSON string
        if not isinstance(condition['value'], str):
            print("[moesif] The field value should be a JSON string")
            return False

        path = condition['path']
        if not isinstance(path, str):
            print("[moesif] The field value should be a JSON string")
            return False

        start = path.find('request.body.')
        if start == -1:
            return False
        start += len('request.body.')
        if '.' in path[start:]:
            print("[moesif] nested fields are not supported for request body condition matching")
            return False

        return True

    def get_updated_response_with_matched_rules(self, governance_rule, rule_and_values):
        """
        get updated response if the governance is blocked checked and matched with request
        :param governance_rule:
        :param rule_and_values:
        :return:
        """
        gr_status, gr_header, gr_body = self.fetch_governance_rule_response_details(governance_rule)

        # Updated governance rule headers
        updated_gr_headers = {}
        if gr_header:
            for k, v in gr_header.items():
                updated_gr_headers[k] = v

        updated_gr_body = None
        if isinstance(gr_body, str):
            updated_gr_body = gr_body
        elif isinstance(gr_body, dict):
            updated_gr_body = gr_body.copy()

        updated_gr_values = {}
        if 'values' in rule_and_values and rule_and_values['values']:
            rule_values = rule_and_values['values']
            for k, v in rule_values.items():
                try:
                    updated_gr_values[int(k)] = v
                except Exception as e:
                    print('[moesif] Error when converting entity rules values key: ', e)

            updated_gr_headers = self.transform_values(updated_gr_headers, updated_gr_values)
            updated_gr_body = self.transform_values(updated_gr_body, updated_gr_values)

        return gr_status, updated_gr_headers, updated_gr_body

    def block_request_based_on_entity_governance_rule(self,
                                                      request_mapping_for_regex_config,
                                                      ready_for_body_request,
                                                      governance_rules,
                                                      entity_rules,
                                                      rule_entity_type,
                                                      entity_id,
                                                      DEBUG):
        """
        Check if need to block request based on the governance rule of the entity associated with the request
        :param request_mapping_for_regex_config:
        :param ready_for_body_request:
        :param governance_rules:
        :param entity_rules:
        :param rule_entity_type:
        :param entity_id:
        :param DEBUG:
        :return: object of updated response status, headers and body, if criteria is matched and block is true, otherwise return None
        """
        response_buffer = BlockResponseBufferList()

        entity_id_rules_mapping = None

        try:
            entity_id_rules_mapping = entity_rules[rule_entity_type][entity_id]
        except KeyError as ke:
            print(
                '[moesif] Skipped blocking request since no governance rules in type of {} with the entity Id - {}: {}'.format(
                    rule_entity_type, entity_id, ke))
        except Exception as e:
            print('[moesif] Skipped blocking request, Error when fetching entity rule with entity {}, {}'.format(
                entity_id, e))

        if not entity_id_rules_mapping:
            return response_buffer

        for rule_and_values in entity_id_rules_mapping:

            try:
                rule_id = rule_and_values['rules']  # rule_id is represented as "rules" in the config schema
            except KeyError as ke:
                print(
                    '[moesif] Skipped a rule in type of {} since the [rule_id] is not found with entity - {}, {}'.format(
                        rule_entity_type, entity_id, ke))
                continue

            governance_rule = governance_rules.get(rule_id, None)

            if not governance_rule or 'response' not in governance_rule or 'status' not in governance_rule['response']:
                if DEBUG:
                    print(
                        '[moesif] Skipped blocking request as governance rule response is not set for the entity Id - ',
                        entity_id)
                continue

            gr_regex_configs = {}
            if "regex_config" in governance_rule and governance_rule["regex_config"]:
                gr_regex_configs = governance_rule["regex_config"]

            matched = not gr_regex_configs or self.check_event_matched_with_governance_rules(
                gr_regex_configs,
                request_mapping_for_regex_config,
                ready_for_body_request)

            if not matched:
                if DEBUG:
                    print(
                        "[moesif] Skipped blocking request as governance rule {} regex conditions does not match".format(
                            rule_id))
                continue

            # update response status, headers and body if one block rule matched
            updated_gr_status, updated_gr_headers, updated_gr_body = self.get_updated_response_with_matched_rules(
                governance_rule, rule_and_values)

            block = governance_rule.get('block', False)
            response_buffer.update(block, updated_gr_status, updated_gr_headers, updated_gr_body)

            if DEBUG:
                print("[moesif] request matched with rule_id [{}]".format(rule_id))

        return response_buffer

    def get_rules_id_if_governance_rule_matched(self, regex_governance_rules, event, ready_for_body_request):
        """
        find the regex governance rules what matched with request, and return the governance rules id
        :param regex_governance_rules:
        :param event:
        :param ready_for_body_request:
        :return:
        """
        matched_rules_id = []
        request_config_mapping = self.prepare_request_config_based_on_regex_config(event, ready_for_body_request)
        for id, rule in regex_governance_rules.items():
            if 'regex_config' not in rule or not rule['regex_config']:
                continue
            regex_configs = rule['regex_config']

            matched = self.check_request_with_regex_match(regex_configs, request_config_mapping,
                                                          ready_for_body_request)

            if matched:
                try:
                    matched_rules_id.append(rule['_id'])
                except KeyError as ke:
                    print('[moesif] Error when fetching matched regex governance rule ', ke)

        return matched_rules_id

    def block_request_based_on_governance_rule_regex_config(self, event, ready_for_body_request, regex_governance_rules,
                                                            DEBUG):
        """
        Check if need to block request based on the governance rule regex config associated with the request
        :param event:
        :param ready_for_body_request:
        :param regex_governance_rules:
        :param DEBUG:
        :return:
        """

        response_buffer = BlockResponseBufferList()
        matched_rules_id = self.get_rules_id_if_governance_rule_matched(regex_governance_rules, event,
                                                                        ready_for_body_request)

        if not matched_rules_id:
            if DEBUG:
                print('[moesif] no regex rule matched with the request')
        else:
            for rule_id in matched_rules_id:
                governance_rule = regex_governance_rules.get(rule_id)
                if not governance_rule:
                    if DEBUG:
                        print(
                            '[moesif] Skipped blocking request as rule {} is not found'.format(rule_id))
                        continue

                if 'response' not in governance_rule \
                        or 'status' not in governance_rule['response'] \
                        or 'headers' not in governance_rule['response']:
                    if DEBUG:
                        print(
                            '[moesif] Skipped blocking request as response is not set for the governance rule with regex config')
                    continue

                block = governance_rule.get('block', False)
                gr_status, gr_header, gr_body = self.fetch_governance_rule_response_details(governance_rule)

                response_buffer.update(block, gr_status, gr_header, gr_body)
                if DEBUG:
                    print('[moesif] request matched with regex rule with rule_id {}'.format(rule_id))
        return response_buffer

    # TODO can deal with request.body in one place
    @classmethod
    def prepare_request_config_based_on_regex_config(cls, event: EventModel, ready_for_body_request):
        """
        fetch regex config from config mapping
        :param event:
        :param ready_for_body_request:
        :return:
        """
        regex_config = {}
        # Config mapping for request.verb
        if event.request.verb:
            regex_config["request.verb"] = event.request.verb

        # Config mapping for request.uri
        if event.request.uri:
            extracted = re.match("http[s]*://[^/]+(/[^?]+)", event.request.uri)
            extracted_string = None
            if not extracted:
                extracted_string = '/'
            else:
                extracted_string = extracted.string
            regex_config["request.route"] = extracted_string

        # Config mapping for request.ip_address
        if event.request.ip_address:
            regex_config["request.ip_address"] = event.request.ip_address

        # DONE Add GraphQL Operation Name
        if event.request.body:
            if 'operationName' in event.request.body:
                regex_config['request.body.operationName'] = event.request.body['operationName']
            if isinstance(event.request.body, dict) and ready_for_body_request:
                if 'query' in event.request.body:
                    regex_config['request.body.query'] = event.request.body['query']
            elif isinstance(event.request.body, str) and ready_for_body_request:
                regex_config['request.body.query'] = base64.b64decode(event.request.body).decode('utf-8')

        # Config mapping for response.status
        if event.response.status:
            regex_config['response.status'] = event.response.status

        return regex_config

    def generate_blocking_response(self, response_buffers):
        """
        rearrange matching rules' response, merge all of the headers, and ordered by the rules type priority
        updated response_body and response_status with the highest priority blocked rule

        :param response_buffers:
        :return: None if none of the rules is blocking request, otherwise return the blocking response
        """
        if not self.check_if_request_blocked(response_buffers):
            return None

        updated_body = None
        updated_status = None
        updated_headers = {}

        for rule_type in REVERSED_PRIORITY_RULES_ORDER:
            if rule_type in response_buffers:
                buffer = response_buffers.get(rule_type, None)
                if buffer:
                    responses = buffer.responses
                    for response in responses:
                        if response.blocked:
                            updated_body = response.block_response_body
                            updated_status = response.block_response_status
                        updated_headers.update(response.block_response_headers)

        gov_rule_response = GovernanceRuleBlockResponse()
        gov_rule_response.update_response(updated_status, updated_headers, updated_body, True)
        return gov_rule_response

    @classmethod
    def check_if_request_blocked(cls, response_buffers):
        for rule_type in REVERSED_PRIORITY_RULES_ORDER:
            if rule_type in response_buffers:
                response = response_buffers[rule_type]
                if response.blocked:
                    return True
        return False

    def govern_request(self,
                       event,
                       user_id, company_id,
                       req_body_transfer_encoding,
                       entity_rules,
                       user_governance_rules,
                       company_governance_rules,
                       regex_governance_rules,
                       DEBUG):
        user_id_entity = user_id
        company_id_entity = company_id

        ready_for_body_request = self.ok_request_body_regex_rule(event.request, req_body_transfer_encoding)
        request_mapping_for_regex_config = self.prepare_request_config_based_on_regex_config(event,
                                                                                             ready_for_body_request)

        response_buffers = {}
        if regex_governance_rules:
            regex_response_buffer = self.block_request_based_on_governance_rule_regex_config(event,
                                                                                             ready_for_body_request,
                                                                                             regex_governance_rules,
                                                                                             DEBUG)
            if not regex_response_buffer.blocked:
                if DEBUG:
                    print('[moesif] No matching with the request from regex rules')

            response_buffers['regex'] = regex_response_buffer
        else:
            if DEBUG:
                print('[moesif] No regex rules')

        if company_id_entity and company_governance_rules:
            company_response_buffer = self.block_request_based_on_entity_governance_rule(
                request_mapping_for_regex_config,
                ready_for_body_request,
                company_governance_rules,
                entity_rules,
                'company_rules',
                company_id_entity,
                DEBUG)
            if not company_response_buffer.blocked:
                if DEBUG:
                    print('[moesif] No blocking from company: ', company_id_entity)

            response_buffers['company'] = company_response_buffer
        else:
            if DEBUG:
                print('[moesif] company_id is not valid or no governance rules for the company')

        if user_id_entity and user_governance_rules:
            user_response_buffer = self.block_request_based_on_entity_governance_rule(request_mapping_for_regex_config,
                                                                                      ready_for_body_request,
                                                                                      user_governance_rules,
                                                                                      entity_rules,
                                                                                      'user_rules',
                                                                                      user_id_entity,
                                                                                      DEBUG)

            if not user_response_buffer.blocked:
                if DEBUG:
                    print('[moesif] No blocking from user: ', user_id_entity)

            response_buffers['user'] = user_response_buffer
        else:
            if DEBUG:
                print('[moesif] user_id is not valid or no governance rules for the user')

        blocking_response = self.generate_blocking_response(response_buffers)

        return blocking_response



import json
from moesifapi import APIException


class GovernanceRulesCacher:

    def __init__(self, api_client):
        self.api_client = api_client
        self.user_rules = {}
        self.company_rules = {}
        self.regex_rules = {}

    def get_governance_rules_from_client(self, DEBUG):
        try:
            get_rules_response = self.api_client.get_governance_rules()
            rules = json.loads(get_rules_response.raw_body)
            return rules
        except APIException as inst:
            if 401 <= inst.response_code <= 403:
                print("[moesif] Unauthorized access getting application configuration. Please check your Appplication Id.")
            if DEBUG:
                print("[moesif] Error getting governance rules, with status code:", inst.response_code)
            return None

        except Exception as ex:
            if DEBUG:
                print("[moesif] Error getting governance rules:", ex)
            return None

    def generate_rules_caching(self, DEBUG):
        try:
            governance_rules = self.get_governance_rules_from_client(DEBUG)
            if not governance_rules:
                return None, None, None
            rule_types = ['regex', 'user', 'company']
            rules_type_mapping = {}
            for rule_type in rule_types:
                rules_type_mapping[rule_type] = {}
            for rule in governance_rules:
                rule_id = rule['_id']

                if 'type' in rule:
                    rule_type = rule['type']

                    if rule_type in rule_types:
                        rules_type_mapping[rule_type][rule_id] = rule
                    else:
                        print('[moesif] Get parsed rule type {} is not valid'.format(rule['type']))

                    self.user_rules = rules_type_mapping['user']
                    self.company_rules = rules_type_mapping['company']
                    self.regex_rules = rules_type_mapping['regex']
        except Exception as e:
            print("[moesif] Error when parsing rules response: ", e)

        return self.user_rules, self.company_rules, self.regex_rules



from moesifapi.models import *
from moesifapi.exceptions.api_exception import *
from moesifapi.api_helper import *


class User:

    def __init__(self):
        pass

    @classmethod
    def update_user(cls, user_profile, api_client, DEBUG):
        if not user_profile:
            print('Expecting the input to be either of the type - UserModel, dict or json while updating user')
        else:
            if isinstance(user_profile, dict):
                if 'user_id' in user_profile:
                    try:
                        api_client.update_user(UserModel.from_dictionary(user_profile))
                        if DEBUG:
                            print('User Profile updated successfully')
                    except APIException as inst:
                        if 401 <= inst.response_code <= 403:
                            print("Unauthorized access sending event to Moesif. Please check your Appplication Id.")
                        if DEBUG:
                            print("Error while updating user, with status code:")
                            print(inst.response_code)
                else:
                    print('To update an user, an user_id field is required')

            elif isinstance(user_profile, UserModel):
                if user_profile.user_id is not None:
                    try:
                        api_client.update_user(user_profile)
                        if DEBUG:
                            print('User Profile updated successfully')
                    except APIException as inst:
                        if 401 <= inst.response_code <= 403:
                            print("Unauthorized access sending event to Moesif. Please check your Appplication Id.")
                        if DEBUG:
                            print("Error while updating user, with status code:")
                            print(inst.response_code)
                else:
                    print('To update an user, an user_id field is required')
            else:
                try:
                    user_profile_json = APIHelper.json_deserialize(user_profile)
                    if 'user_id' in user_profile_json:
                        try:
                            api_client.update_user(UserModel.from_dictionary(user_profile_json))
                            if DEBUG:
                                print('User Profile updated successfully')
                        except APIException as inst:
                            if 401 <= inst.response_code <= 403:
                                print("Unauthorized access sending event to Moesif. Please check your Appplication Id.")
                            if DEBUG:
                                print("Error while updating user, with status code:")
                                print(inst.response_code)
                    else:
                        print('To update an user, an user_id field is required')
                except:
                    print('Error while deserializing the json, please make sure the json is valid')

    @classmethod
    def update_users_batch(cls, user_profiles, api_client, DEBUG):
        if not user_profiles:
            print('Expecting the input to be either of the type - List of UserModel, dict or json while updating users')
        else:
            if all(isinstance(user, dict) for user in user_profiles):
                if all('user_id' in user for user in user_profiles):
                    try:
                        batch_profiles = [UserModel.from_dictionary(d) for d in user_profiles]
                        api_client.update_users_batch(batch_profiles)
                        if DEBUG:
                            print('Users Profile updated successfully')
                    except APIException as inst:
                        if 401 <= inst.response_code <= 403:
                            print("Unauthorized access sending event to Moesif. Please check your Appplication Id.")
                        if DEBUG:
                            print("Error while updating users, with status code:")
                            print(inst.response_code)
                else:
                    print('To update users, an user_id field is required')

            elif all(isinstance(user, UserModel) for user in user_profiles):
                if all(user.user_id is not None for user in user_profiles):
                    try:
                        api_client.update_users_batch(user_profiles)
                        if DEBUG:
                            print('Users Profile updated successfully')
                    except APIException as inst:
                        if 401 <= inst.response_code <= 403:
                            print("Unauthorized access sending event to Moesif. Please check your Appplication Id.")
                        if DEBUG:
                            print("Error while updating users, with status code:")
                            print(inst.response_code)
                else:
                    print('To update users, an user_id field is required')
            else:
                try:
                    user_profiles_json = [APIHelper.json_deserialize(d) for d in user_profiles]
                    if all(isinstance(user, dict) for user in user_profiles_json) and all(
                                    'user_id' in user for user in user_profiles_json):
                        try:
                            batch_profiles = [UserModel.from_dictionary(d) for d in user_profiles_json]
                            api_client.update_users_batch(batch_profiles)
                            if DEBUG:
                                print('Users Profile updated successfully')
                        except APIException as inst:
                            if 401 <= inst.response_code <= 403:
                                print("Unauthorized access sending event to Moesif. Please check your Appplication Id.")
                            if DEBUG:
                                print("Error while updating users, with status code:")
                                print(inst.response_code)
                    else:
                        print('To update users, an user_id field is required')
                except:
                    print('Error while deserializing the json, please make sure the json is valid')
